/**
 * Define api of apeman-api-recover
 * @function define
 * @param {Object} [options] - Optional settings
 * @param {boolean} [options.logging] - Enable logger
 * @returns {Object} - API configuration
 */
'use strict'

const co = require('co')
const assert = require('assert')
const uuid = require('uuid')
const aslogger = require('aslogger')
const apemanApiSign = require('apeman-api-sign')
const spec = require('./spec')
const load = require('apeman-api-client/load')
const { recoverNotFoundError, recoverTokenWrongError, recoverTokenExpiredError } = require('apemanerror/buildin')

/** @lends define */
function define (User, UserSign, UserRecover, options = {}) {
  assert.ok(UserRecover, 'UserRecover is required.')
  assert.ok(!!(UserRecover.$$inherited && UserRecover.$$inherited.ApUserRecover), 'UserRecover should inherit ApUserRecover.')
  let { passwordSpec, logging } = options
  let logger = aslogger({ disabled: !logging })

  let signApi = apemanApiSign(User, UserSign, {
    passwordSpec, logging
  })

  return {
    /**
     * Test the reachability of the api.
     * @method ping
     * @param {string} [pong='pong'] - Pong message to return
     * @returns {string} - Pong message
     */
    ping (pong = 'pong') {
      return co(function * () {
        logger.trace('pong:', pong)
        return pong
      })
    },

    /**
     * Issue recover token to email
     * @method issue
     * @param {string} email
     * @returns {Promise}
     */
    issue (email) {
      let {
        expireDuration = 26 * 60 * 60 * 10000
      } = options
      return co(function * () {
        let user = email && (yield User.findOne({ where: { email } }))
        if (!user) {
          throw recoverNotFoundError()
        }
        let token = uuid.v4().replace(/\-/g, '')
        let userId = user.id
        let [recover] = (yield UserRecover.findOrCreate({
          where: { userId, email },
          defaults: { userId, token, email }
        }))
        return yield recover.update({
          token,
          expireAt: new Date(new Date().getTime() + expireDuration)
        })
      })
    },

    /**
     * Confirm email
     * @method check
     * @param {string} token - Recover token
     * @returns {Promise}
     */
    check (token) {
      const s = this
      return co(function * () {
        let recover = yield s._getRecover(token)
        return !!recover
      })
    },

    /**
     * Update password
     * @method update
     * @param {string} token - Recover token
     * @param {string} newPassword - New password
     * @returns {Promise}
     */
    update (token, newPassword) {
      const s = this
      let { session } = s
      return co(function * () {
        let sign = yield load(signApi, { $$ctx: s })
        let recover = yield s._getRecover(token)
        let { userId } = recover
        let userSigns = (yield UserSign.findAll({
          where: { userId },
          include: [ UserSign.getRef('user') ]
        }))
        let userSign = userSigns[ 0 ]
        session.signed = apemanApiSign.formatSigned(userSign.user, userSign)
        let keys = userSigns.reduce((keys, userSign) => Object.assign(keys, {
          [userSign.kind]: userSign.key
        }), {})
        yield sign.signmod(keys, newPassword)
        yield sign.signout()
        yield recover.update({
          doneAt: new Date()
        })
        return { key: userSign.key, password: newPassword }
      })
    },

    /**
     * Get an recover
     * @method one
     * @param {string} id
     * @returns {Promise}
     */
    one (id){
      return UserRecover.findById(id)
    },

    /**
     * Get last recover for current user
     * @method last
     * @param {string} email
     * @returns {Promise}
     */
    last (email) {
      return co(function * () {
        let user = email && (yield User.findOne({ where: { email } }))
        if (!user) {
          throw recoverNotFoundError()
        }
        let { id: userId } = user
        let recovers = yield UserRecover.findAll({
          where: { userId },
          limit: 1,
          order: [
            [ 'expireAt', 'DESC' ]
          ]
        })
        return recovers.shift()
      })
    },

    /**
     * Get recover from token
     * @param {string} token - Token to search
     * @throws {Error} - Unless valid token found
     * @returns {Promise}
     */
    _getRecover (token) {
      return co(function * () {
        let recover = token && (yield UserRecover.findOne({ token }))
        if (!recover) {
          throw recoverTokenWrongError()
        }
        if (recover.expired) {
          throw recoverTokenExpiredError()
        }
        if (recover.done) {
          throw recoverTokenExpiredError()
        }
        return recover
      })
    },

    /**
     * A hook which runs before method invoked
     * @param {string} methodName - Name of method to invoke
     * @param {Array} params - Prams for invocation
     * @returns {Promise}
     */
    $before (methodName, params) {
      const s = this
      let { state } = s
      return co(function * () {
        state.invokedAt = new Date()
        logger.info(`${methodName} started...`)
      })
    },

    /**
     * A hook which runs after method invoked
     * @param {string} methodName - Name of method to invoke
     * @param {Array} params - Prams for invocation
     * @param {?*} returns - Return values
     * @returns {Promise}
     */
    $after (methodName, params, returns) {
      const s = this
      let { state } = s
      return co(function * () {
        let took = new Date() - state.invokedAt
        logger.info(`...${methodName} done! (${took}ms)`)
      })
    },
    /**
     * Method specifications
     */
    $spec: spec

  }
}

module.exports = define
