/**
 * Test case for define.
 * Runs with mocha.
 */
'use strict'

const define = require('../lib/define.js')
const assert = require('assert')
const asleep = require('asleep')
const { moduleProxy } = require('apemanmock')
const co = require('co')
const fs = require('fs')
const apemanApiSign = require('apeman-api-sign')
const apemandb = require('apemandb')
const apemanmodel = require('apemanmodel')
const { SAMPLE_DEFS_DIR } = apemanmodel.paths

describe('define', function () {
  this.timeout(5000)
  let tmpDir = `${__dirname}/../tmp`
  let dbFile = `${tmpDir}/testing-db.db`
  let db, Session, User, UserSign, UserRecover

  before(() => co(function * () {
    db = apemandb({
      env: {
        DATABASE: 'testing-database01',
        USERNAME: 'foo',
        PASSWORD: 'bar',
        DIALECT: 'sqlite',
        STORAGE: dbFile
      },
      models: `${SAMPLE_DEFS_DIR}/*.json`
    })
    Session = db.model('Session')
    User = db.model('User')
    UserSign = db.model('UserSign')
    UserRecover = db.model('UserRecover')
    yield db.sync({ force: true })
  }))

  after(() => co(function * () {
    yield asleep(10) // Wait for flush.
    fs.unlinkSync(dbFile)
  }))

  it('Create an instance', () => co(function * () {
    let module = define(User, UserSign, UserRecover, {
      logging: true
    })
    let api = moduleProxy(module, { ctx: { state: {} } })
    assert.ok(api)
  }))

  it('Call API methods', () => co(function * () {
    let ctx = { state: {}, session: {} }

    let recover = moduleProxy(define(User, UserSign, UserRecover, {
      logging: true
    }), { ctx })

    let pong = yield recover.ping('The pong!')
    assert.equal(pong, 'The pong!')

    {
      let sign = moduleProxy(apemanApiSign(User, UserSign, { logging: false }), { ctx })
      let email = 'foo@example.com'
      let signed = yield sign.signup('foo', email, 'foo1234')
      assert.equal(signed.key, 'foo')
      assert.equal(signed.email, email)
      assert.ok(signed.sign.hash)
      yield sign.signout()

      let { token, userId } = yield recover.issue(email)

      assert.ok(yield recover.last(email))
      assert.ok(token)
      assert.ok(userId)
      assert.ok(yield recover.check(token))
      yield recover.update(token, 'hogehoge')
      let recovered = yield sign.signget()
    }
  }))

  it('Compare spec and implement', () => co(function * () {
    let module = define(User, UserSign, UserRecover, {
      logging: false
    })
    let implNames = Object.keys(module).filter((name) => !/^[_\$]/.test(name))
    let specNames = Object.keys(module.$spec)
    for (let implName of implNames) {
      assert.ok(!!~specNames.indexOf(implName), `${implName} is implemented, but not described in "$spec"`)
    }
    for (let specName of specNames) {
      assert.ok(!!~implNames.indexOf(specName), `${specName} is described in "$spec", but not implemented`)
    }
  }))
})

/* global describe, before, after, it */
